#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 21 15:23:44 2021

@author: yuemingzhuo
"""

import torch
from torch.nn import Conv2d, BatchNorm2d, ReLU, AvgPool2d, MaxPool2d, UpsamplingNearest2d, Sigmoid

class ConvBlock(torch.nn.Module):
    def __init__(self, in_channels=1, growth_rate=4):
        super(ConvBlock, self).__init__()
        
        layers = []
        
        layers.append(BatchNorm2d(in_channels))
        layers.append(ReLU(inplace=True))
        layers.append(Conv2d(in_channels, 4*growth_rate, 1, bias=False))
        layers.append(BatchNorm2d(4*growth_rate))
        layers.append(ReLU(inplace=True))
        layers.append(Conv2d(4*growth_rate, growth_rate, 3, padding="same", bias=False))
        
        self.net = torch.nn.Sequential(*layers)
        
        
    def forward(self,x):
        return torch.cat((x, self.net(x)), axis=1)
    
    
    
    
class DenseBlock(torch.nn.Module):
    def __init__(self, in_channels=32, N_blocks=6, growth_rate=4):
        super(DenseBlock, self).__init__()
        
        blocks = []
        Cin = in_channels
        
        for i in range(N_blocks):
            blocks.append(ConvBlock(in_channels=Cin, growth_rate=growth_rate))
            Cin += growth_rate
            
        self.net = torch.nn.Sequential(*blocks)
        
        
    def forward(self,x):
        return self.net(x)

    
    
    
class TransitionBlock(torch.nn.Module):
    def __init__(self, in_channels=32, reduction=0.5):
        super(TransitionBlock, self).__init__()
        
        layers = []
        
        layers.append(BatchNorm2d(in_channels))
        layers.append(ReLU(inplace=True))
        layers.append(Conv2d(in_channels, int(in_channels*reduction), 1, bias=False))
    
        self.net1 = torch.nn.Sequential(*layers)
        
        layers.append(AvgPool2d(2, stride=2))
        
        self.net2 = torch.nn.Sequential(*layers)
        
    def forward(self,x):
        return self.net2(x), self.net1(x)
    

    
    
class DecoderBlock(torch.nn.Module):
    def __init__(self, in_channels=32, growth_rate=4):
        super(DecoderBlock, self).__init__()
        
        layers = []
        
        layers.append(Conv2d(in_channels, growth_rate, 3, padding= "same", bias=False))
        layers.append(BatchNorm2d(growth_rate))
        layers.append(ReLU(inplace=True))
        layers.append(Conv2d(growth_rate, growth_rate, 3, padding= "same", bias=False))
        layers.append(BatchNorm2d(growth_rate))
        layers.append(ReLU(inplace=True))
        
        self.net = torch.nn.Sequential(*layers)
        
    def forward(self, x, x1):
        x2 = torch.cat((x,x1), axis=1)
        return self.net(x2)
    
    
    
    
    
class SuperDenseNet(torch.nn.Module):
    def __init__(self, in_channels=1, N_ConvBlocks=(6,12,24,16)):
        super(SuperDenseNet, self).__init__()
        
        # 1st block: Conv 7x7 -> output size: (N,64,256,256)
        self.Conv1 = Conv2d(in_channels, 64, 7, stride=2, padding=3, bias=False)
        self.BN1 = BatchNorm2d(64)
        self.ReLU1 = ReLU(inplace=True)
        
        # 2nd block: Maxpooling (2x2) -> output size: (N,64,128,128)
        self.MaxPool1 = MaxPool2d(3, stride=2, padding=1)
        
        # 3rd block: Denseblock1 (6x ConvBlocks) -> output size: (N,88,128,128)
        self.DB1 = DenseBlock(in_channels=64, N_blocks=N_ConvBlocks[0])
        
        # 4th block: TransitionBlock1
        # 2 outputs: x1->(N,44,64,64), y1->(N,44,128,128)
        self.TB1 = TransitionBlock(in_channels=88, reduction=0.5)
        
        # 5th block: Denseblock2 (12x ConvBlocks) -> output size: (N,92,64,64)
        self.DB2 = DenseBlock(in_channels=44, N_blocks=N_ConvBlocks[1])
        
        # 6th block: TransitionBlock2
        # 2 outputs: x2->(N,46,32,32), y2->(N,46,64,64)
        self.TB2 = TransitionBlock(in_channels=92, reduction=0.5)
        
        # 7th block: Denseblock3 (24x ConvBlocks) -> output size: (N,142,32,32)
        self.DB3 = DenseBlock(in_channels=46, N_blocks=N_ConvBlocks[2])
        
        # 8th block: TransitionBlock3
        # 2 outputs: x3->(N,71,16,16), y3->(N,71,32,32)
        self.TB3 = TransitionBlock(in_channels=142, reduction=0.5)
        
        # 9th block: DenseBlock4 (16x ConvBlocks) -> output size: (N,135,16,16)
        self.DB4 = DenseBlock(in_channels=71, N_blocks=N_ConvBlocks[3])
        
        # 10th block:
        self.BN2 = BatchNorm2d(135)
        self.ReLU2 = ReLU(inplace=True)
        
        # 11th block: Upsampling (2x2) -> output size: (N,135,32,32)
        self.US1 = UpsamplingNearest2d(scale_factor=2)
        
        # 12th block: DecoderBlock1
        # Input channel: 71 + 135 = 206 -> output size: (N,256,32,32)
        self.DCB1 = DecoderBlock(in_channels=206, growth_rate=256)
        
        # 13th block: Upsampling (2x2) -> output size: (N,256,64,64)
        self.US2 = UpsamplingNearest2d(scale_factor=2)
        
        # 14th block: DecoderBlock2
        # Input channel: 46 + 256 = 302 -> output size: (N,128,64,64)
        self.DCB2 = DecoderBlock(in_channels=302, growth_rate=128)
        
        # 15th block: Upsampling (2x2) -> output size: (N,128,128,128)
        self.US3 = UpsamplingNearest2d(scale_factor=2)
        
        # 16th block: DecoderBlock3
        # Input channel: 44 + 128 = 172 -> output size: (N,64,128,128)
        self.DCB3 = DecoderBlock(in_channels=172, growth_rate=64)
        
        # 17th block: Upsampling (2x2) -> output size: (N,64,256,256)
        self.US4 = UpsamplingNearest2d(scale_factor=2)
        
        # 18th block: DecoderBlock3
        # Input channel: 64 + 64 = 128 -> output size: (N,32,256,256)
        self.DCB4 = DecoderBlock(in_channels=128, growth_rate=32)
        
        # 19th block: Upsampling (2x2) -> output size: (N,32,512,512)
        self.US5 = UpsamplingNearest2d(scale_factor=2)
        
        # 20th block: Conv (3x3) -> output size: (N,16,512,512)
        self.Conv2 = Conv2d(32, 16, 3, padding="same", bias=False)
        self.BN3 = BatchNorm2d(16)
        self.ReLU3 = ReLU(inplace=True)
        
        # 21th block: Conv (3x3) -> output size: (N,16,512,512)
        self.Conv3 = Conv2d(16, 16, 3, padding="same", bias=False)
        self.BN4 = BatchNorm2d(16)
        self.ReLU4 = ReLU(inplace=True)
        
        # 22th block: Conv (1x1) -> output size: (N,1,512,512) -> Compensated OCT image
        self.Conv4 = Conv2d(16, 1, 1, padding="same", bias=True)
        self.Sig = Sigmoid()
        
        
        
    def forward(self, x):

        y0 = self.ReLU1(self.BN1(self.Conv1(x))) # 1st  block (conv)
        y = self.MaxPool1(y0)                    # 2nd  block (pooling)
        y = self.DB1(y)                          # 3rd  block (dense1)
        x1, y1 = self.TB1(y)                     # 4th  block (transition1)
        y = self.DB2(x1)                         # 5th  block (dense2)
        x2, y2 = self.TB2(y)                     # 6th  block (transition2)
        y = self.DB3(x2)                         # 7th  block (dense3)
        x3, y3 = self.TB3(y)                     # 8th  block (transition3)
        y = self.DB4(x3)                         # 9th  block (dense4)
        y = self.ReLU2(self.BN2(y))              # 10th block (nonlinear)
        y = self.US1(y)                          # 11th block (upsampling1)
        y = self.DCB1(y,y3)                      # 12th block (decoder1)
        y = self.US2(y)                          # 13th block (upsampling2)
        y = self.DCB2(y,y2)                      # 14th block (decoder2)
        y = self.US3(y)                          # 15th block (upsampling3)
        y = self.DCB3(y,y1)                      # 16th block (decoder3)
        y = self.US4(y)                          # 17th block (upsampling4)
        y = self.DCB4(y,y0)                      # 18th block (decoder4)
        y = self.US5(y)                          # 19th block (upsampling5)
        y = self.ReLU3(self.BN3(self.Conv2(y)))  # 20th block (conv)
        y = self.ReLU4(self.BN4(self.Conv3(y)))  # 21st block (conv)
        y = self.Sig(self.Conv4(y))              # 22nd block (output)
        
        return y